/*
 * IPVS:        Round-Robin Scheduling module
 *
 * Version:     $Id: ip_vs_rr.c,v 1.3 1999/09/08 03:30:55 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@iinchina.net>
 *              Peter Kese <peter.kese@ijs.si>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Fixes/Changes:
 *     Wensong Zhang            :     changed the ip_vs_rr_schedule to return dest
 *     Julian Anastasov         :     fixed the NULL pointer access bug in debugging
 *     Wensong Zhang            :     changed some comestics things for debugging
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <net/ip_masq.h>
#ifdef CONFIG_IP_MASQUERADE_MOD
#include <net/ip_masq_mod.h>
#endif
#include <linux/sysctl.h>
#include <linux/ip_fw.h>
#include <net/ip_vs.h>


static int ip_vs_rr_init_svc(struct ip_vs_service *svc)
{
        MOD_INC_USE_COUNT;
        return 0;
}


static int ip_vs_rr_done_svc(struct ip_vs_service *svc)
{
        MOD_DEC_USE_COUNT;
        return 0;
}


/*
 * Round-Robin Scheduling
 */
static struct ip_vs_dest* ip_vs_rr_schedule(struct ip_vs_service *svc)
{
	IP_VS_DBG("ip_vs_rr_schedule(): Scheduling...\n");

        if (svc->sched_data != NULL) 
                svc->sched_data = ((struct ip_vs_dest*)svc->sched_data)->next;
        if (svc->sched_data == NULL) 
                svc->sched_data = svc->destinations;

#ifdef CONFIG_IP_VS_DEBUG
        {
                struct ip_vs_dest *dest;

                dest=(struct ip_vs_dest*)svc->sched_data;

                if (dest != NULL) {
                        IP_VS_DBG("RR: server %d.%d.%d.%d:%d "
                                  "activeconns %d refcnt %d weight %d\n",
                                  NIPQUAD(dest->addr), ntohs(dest->port),
                                  atomic_read(&dest->activeconns),
                                  atomic_read(&dest->refcnt), dest->weight);
                }
        }
#endif

	return (struct ip_vs_dest*)svc->sched_data;
}


static struct ip_vs_scheduler ip_vs_rr_scheduler = {
	NULL,			/* next */
	"rr",			/* name */
	ATOMIC_INIT(0),		/* refcnt */
	ip_vs_rr_init_svc,      /* service initializer */
	ip_vs_rr_done_svc,      /* service done */
	ip_vs_rr_schedule,      /* select a server and create new masq entry */
};


__initfunc(int ip_vs_rr_init(void))
{
	IP_VS_INFO("Initializing RR scheduling\n");
	return register_ip_vs_scheduler(&ip_vs_rr_scheduler) ;
}


#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
	/* module initialization by 'request_module' */
	if(register_ip_vs_scheduler(&ip_vs_rr_scheduler) != 0)
	        return -EIO;

	IP_VS_INFO("RR scheduling module loaded.\n");
	
        return 0;
}

void cleanup_module(void)
{
	/* module cleanup by 'release_module' */
	if(unregister_ip_vs_scheduler(&ip_vs_rr_scheduler) != 0)
	        IP_VS_INFO("cannot remove RR scheduling module\n");
	else
	        IP_VS_INFO("RR scheduling module unloaded.\n");
}

#endif /* MODULE */
